package gmap;

import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Rect;
import android.graphics.Bitmap.CompressFormat;
import android.graphics.Bitmap.Config;

public class GMapHandler {
	private static Map<String, byte[]> gmapPng = new HashMap<String, byte[]>();
	private static Map<String, Bitmap> gmapBmp = new HashMap<String, Bitmap>();
	private static Thread compressThread = null;
	private static Thread downloadThread = null;
	private static List<GPoint> downloadIndex = new ArrayList<GPoint>();
	Bitmap buffer;
	
	//wľߪު
	String bufferIndex;
	Bitmap loadimg = null;
	GPoint center;
	int width;
	int height;
	String apiKey;
	DownloadCallback callBack = new DownloadCallback() {
		public void downloadCallback() {

		}
	};

	public GMapHandler(GLatLng centerLatLng, int zoom, int width, int height,
			String apiKey) {
		center = new GPoint(centerLatLng, zoom);
		inti(center, width, height, apiKey);

	}

	public GMapHandler(GPoint center, int width, int height, String apiKey) {
		inti(center, width, height, apiKey);
	}

	private void inti(GPoint center, int width, int height, String apiKey) {
		this.center = center;
		this.width = width;
		this.height = height;
		this.apiKey = apiKey;
		int drawCountX = (width / 256) + 2;
		int drawCountY = (height / 256) + 2;
		buffer = Bitmap.createBitmap(drawCountX * 256, drawCountY * 256,
				Config.ARGB_8888);

	}

	public void setLoadingImage(Resources res, int DrawableId) {
		loadimg = BitmapFactory.decodeResource(res, DrawableId);
	}

	public void setDownloadCallBack(GMapHandler.DownloadCallback callback) {
		this.callBack = callback;
	}

	public void draw(Canvas canvas) {
		updateBuffer();
		Paint paint = new Paint();
		paint.setAlpha(255);
		GPoint drawIndex = new GPoint(center);
		drawIndex.offset(-width / 2, -height / 2);
		int offsetX = drawIndex.X % 256;
		int offsetY = drawIndex.Y % 256;
		Rect src = new Rect(offsetX, offsetY, offsetX + width, offsetY + height);
		Rect dst = new Rect(0, 0, width, height);
		canvas.drawBitmap(buffer, src, dst, paint);
	}

	public Bitmap getBitmap() {
		updateBuffer();
		return Bitmap.createBitmap(buffer);
	}

	private Bitmap getLoadimg() {
		if (loadimg == null)
			loadimg = Bitmap.createBitmap(width, height, Config.ARGB_8888);
		return loadimg;

	}

	private void updateBuffer() {
		compressGmpBmp();
		if (bufferIndex != getIndexString(center)) {
			bufferIndex = getIndexString(center);
			Canvas canvas = new Canvas(buffer);
			Paint paint = new Paint();
			paint.setAlpha(255);

			GPoint drawIndex = new GPoint(center);
			drawIndex.offset(-width / 2, -height / 2);
			int drawCountX = (width / 256) + 2;
			int drawCountY = (height / 256) + 2;
			Rect src = new Rect(0, 0, 256, 256);
			for (int iy = 0; iy < drawCountY; iy++) {
				for (int ix = 0; ix < drawCountX; ix++) {
					int left = ix * 256;
					int top = iy * 256;
					int right = left + 256;
					int bottom = top + 256;
					Rect dst = new Rect(left, top, right, bottom);
					Bitmap indexGmap = getIndexGMap(drawIndex);
					canvas.drawBitmap(indexGmap, src, dst, paint);
					drawIndex.offset(256, 0);
				}
				drawIndex.offset(-256 * drawCountX, 256);
			}
		}

	}


	public int getIndexX(GPoint index) {
		return index.X / 256;
	}

	public int getIndexY(GPoint index) {
		return index.Y / 256;
	}

	public String getIndexString(GPoint index) {
		int indexX = getIndexX(index);
		int indexY = getIndexY(index);
		return String.format("IGX=%1$s IGY=%2$s GZ=%3$s", indexX, indexY,
				index.Zoom);
	}

	// oGoogleMap
	public Bitmap getIndexGMap(GPoint index) {
		String indexString = getIndexString(index);

		Bitmap reImg = gmapBmp.get(indexString);
		if (reImg != null) {
			if (!reImg.isRecycled()) {
				return reImg;
			}
		}

		reImg = pngByteToBitmap(gmapPng.get(indexString));
		if (reImg != null) {
			gmapBmp.put(getIndexString(index), reImg);
			return reImg;
		} else {
			gmapBmp.put(indexString, getLoadimg());
			asyncDownlod(index);
			return gmapBmp.get(indexString);
		}

	}


	// YϤ禡
	Runnable compressGmpBmp = new Runnable() {
		public void run() {
			
			int drawCountX = (width / 256) + 2;
			int drawCountY = (height / 256) + 2;
			if (gmapBmp.size() > drawCountX * drawCountY) {
				GPoint drawIndex = new GPoint(center);
				drawIndex.offset(-width / 2, -height / 2);
				
				//M䤣ݭnϥΪϤ
				List<String> needBmpIndex = new ArrayList<String>();
				for (int iy = 0; iy < drawCountY; iy++) {
					for (int ix = 0; ix < drawCountX; ix++) {
						needBmpIndex.add(drawIndex.toString());
						drawIndex.offset(256, 0);
					}
					drawIndex.offset(-256 * drawCountX, 256);
				}

				String[] gmapBmpKey = new String[gmapBmp.size()];
				gmapBmp.keySet().toArray(gmapBmpKey);
				for (String key : gmapBmpKey) {
					if (!needBmpIndex.contains(key)) {
						Bitmap compressBmp = gmapBmp.get(key);

						if (compressBmp != getLoadimg()){
							//xsPNG@֨
							gmapPng.put(key, bitmapToPngByte(compressBmp));
							compressBmp.recycle();
							}
						gmapBmp.remove(key);
						
					}
				}
				//O
				//System.gc();
			}
			
		}
	};

	// YϤ
	public void compressGmpBmp() {

		if (compressThread == null) {
			compressThread = new Thread(compressGmpBmp);
			compressThread.start();
		} else if (!compressThread.isAlive()) {

			compressThread = new Thread(compressGmpBmp);
			compressThread.start();
		}
	}
	
	private Bitmap pngByteToBitmap(byte[] pngByte) {
		if (pngByte != null)
			return BitmapFactory.decodeByteArray(pngByte, 0, pngByte.length);
		else
			return null;
	}

	private byte[] bitmapToPngByte(Bitmap bitmap) {
		if (bitmap != null) {
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			bitmap.compress(CompressFormat.PNG, 0, out);
			byte[] bytes = out.toByteArray();
			return bytes;
		} else {
			return null;
		}
	}

	private void asyncDownlod(GPoint index) {
		downloadIndex.add(new GPoint(index));

		if (downloadThread == null) {
			downloadThread = new Thread(asyncDownlod);
			downloadThread.start();
		} else if (!downloadThread.isAlive()) {
			downloadThread = new Thread(asyncDownlod);
			downloadThread.start();
		}

	}

	public void offset(int dx, int dy) {
		center.offset(dx, dy);
	}

	public void setCenter(GLatLng centerLatLng, int zoom) {
		center = new GPoint(centerLatLng, zoom);
	}

	public void setZoom(int zoom) {
		center.setZoom(zoom);
		updateBuffer();
	}

	public int getZoom() {
		return center.Zoom;
	}

	public interface DownloadCallback {
		public void downloadCallback();
	}

	Runnable asyncDownlod = new Runnable() {
		public void run() {
			while (downloadIndex.size() > 0) {
				GPoint index = downloadIndex.get(0);
				int cx = getIndexX(index) * 256 + 128;
				int cy = getIndexY(index) * 256 + 128;
				GPoint imgCenter = new GPoint(cx, cy, index.Zoom);
				Bitmap downloadImg = GMapDownloader.getNoLogoGMap(imgCenter,
						256, 256, apiKey);
				if (downloadImg == null)
					downloadImg = getLoadimg();
				gmapBmp.put(getIndexString(index), downloadImg);
				downloadIndex.remove(0);
				callBack.downloadCallback();
			}

		}
	};

}
